/*
 * GTimer
 *
 * Copyright:
 *	(C) 1998 Craig Knudsen, cknudsen@radix.net
 *	See accompanying file "COPYING".
 * 
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version 2
 *	of the License, or (at your option) any later version.
 * 
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 * 
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the
 *	Free Software Foundation, Inc., 59 Temple Place,
 *	Suite 330, Boston, MA  02111-1307, USA
 *
 * Description:
 *	Helps you keep track of time spent on different tasks.
 *
 * Author:
 *	Craig Knudsen, cknudsen@radix.net, http://www.radix.net/~cknudsen
 *
 * Home Page:
 *	http://www.radix.net/~cknudsen/gtimer/
 *
 * History:
 *	22-Mar-1998	Added midnight_offset as an parameter to
 *			TaskGetAnnotationEntries.  Entries after midnight
 *			were showing up on the day after the hours were
 *			recorded.
 *	20-Mar-1998	Replaced label widget with a text widget.
 *	18-Mar-1998	Added calls to gtk_window_set_wmclass so the windows
 *			behave better for window managers.
 *	18-Mar-1998	Fixed bug where all weekdays show up as "Mon" in
 *			reports.
 *	18-Mar-1998	Fixed bug where for loop went past array boundaries
 *			and called free on invalid address.
 *	17-Mar-1998	Changed font in report window to be courier so
 *			that text would line up correctly.
 *	17-Mar-1998	Added option of including annotations and make
 *			including hours worked optional.
 *	16-Mar-1998	Removed reference to sys_errlist[] for
 *			compatibility with glibc.
 *	13-Mar-1998	Use midnight_offset when figuring out what the
 *			current date is.
 *	13-Mar-1998	Include seconds in reports
 *	12-Mar-1998	Remove toggle buttons for report type.  Now set
 *			by calling routine.
 *	10-Mar-1998	Added save to file
 *	10-Mar-1998	Added additional report options:
 *			  allow hours for each day to be optional
 *			  include summary for day
 *			  include summary for week
 *			  include summary for month
 *			  include days with no hours
 *	10-Mar-1998	Commented out all tooltips since the API is
 *			different for GTK-0.99.3 and GTK-0.99.5.
 *	08-Mar-1998	Created
 *
 ****************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pwd.h>
#include <time.h>
#include <memory.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <gtk/gtk.h>

#include "task.h"
#include "gtimer.h"

#ifdef GTIMER_MEMDEBUG
#include "memdebug/memdebug.h"
#endif

#define NO_TOOLTIPS	1

#define ONE_DAY	(3600*24)
static int month_days[] = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
static int lmonth_days[] = { 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

extern TaskData **tasks;
extern int num_tasks;
extern int midnight_offset;
extern GtkWidget *main_window;

/* Report types */
#define REPORT_RANGE_TODAY              0
#define REPORT_RANGE_THIS_WEEK          1
#define REPORT_RANGE_LAST_WEEK          2
#define REPORT_RANGE_THIS_MONTH         3
#define REPORT_RANGE_LAST_MONTH         4
#define REPORT_RANGE_THIS_YEAR          5
static char *time_options[] = { "Today", "This Week", "Last Week",
  "This Month", "Last Month", "This Year", NULL };

typedef struct {
  TaskData *taskdata;
  time_t weekly_total;
  time_t monthly_total;
  time_t yearly_total;
  time_t total;
  char *todays_annotations;
  char *weekly_annotations;
  char *monthly_annotations;
  char *yearly_annotations;
  char *total_annotations;
  char week_start[20];
} ReportTaskData;

typedef struct {
  GtkWidget *window;
  GtkWidget *time_menu;
  GtkWidget *time_menu_items[6];
  GtkWidget *include_hours_checkbox;
  GtkWidget *include_annotations_checkbox;
  GtkWidget *task_list;
  TaskData **tasks;
  int num_tasks;
  GtkWidget **list_items;
  report_type type;
  int include_hours;
  int include_annotations;
} ReportData;

typedef struct {
  GtkWidget *window;
  char *text;
  GtkWidget *filesel;
} DisplayReportData;




/*
** Append an annotation to the given pointer after we realloc some
** space for it.  optionally include the date and time and indent
** all lines to make it look nice.
*/
static void concat_annotation ( ptr, annotation, include_date,
  include_time, indentation )
char **ptr;
TaskAnnotation *annotation;
int include_date, include_time;
char *indentation;
{
  struct tm *tm;
  char date_time_str[30];
  char padding[30], *p;
  char *anntext, *newtext;
  int first;

  strcpy ( date_time_str, indentation );
  tm = localtime ( &annotation->text_time );
  if ( include_date )
    sprintf ( date_time_str + strlen ( date_time_str ),
      "%02d/%02d/%02d ", tm->tm_mon + 1,
      tm->tm_mday, tm->tm_year % 100 );

  if ( include_time )
    sprintf ( date_time_str + strlen ( date_time_str ),
      "%02d:%02d ", tm->tm_hour, tm->tm_min );

  strcpy ( padding, date_time_str );
  for ( p = padding; *p != '\0'; p++ )
    *p = ' ';

  /* make a copy of the annotation text because strtok() stomps on it */
  anntext = (char *) malloc ( strlen ( annotation->text ) + 1 );
  strcpy ( anntext, annotation->text );

  if ( *ptr ) {
    newtext = (char *) malloc ( strlen ( *ptr ) + 1 );
    strcpy ( newtext, *ptr );
  } else {
    newtext = (char *) malloc ( 1 );
    newtext[0] = '\0';
  }
  first = 1;
  p = strtok ( anntext, "\n" );
  while ( p ) {
    newtext = (char *) realloc ( newtext,
      strlen ( newtext ) + strlen ( p ) + strlen ( padding ) + 2 );
    if ( first ) {
      first = 0;
      strcat ( newtext, date_time_str );
    } else {
      strcat ( newtext, padding );
    }
    strcat ( newtext, p );
    strcat ( newtext, "\n" );
    p = strtok ( NULL, "\n" );
  }
  free ( anntext );

  if ( *ptr )
    free ( *ptr );
  *ptr = newtext;
}



static void display_results_ok_callback ( widget, data )
GtkWidget *widget;
gpointer data;
{
  DisplayReportData *drd = (DisplayReportData *)data;
  gtk_widget_destroy ( drd->window );
  if ( drd->filesel )
    gtk_widget_destroy ( drd->filesel );
  free ( drd->text );
  free ( drd );
}

static void display_results_save_cancel_callback ( widget, data )
GtkWidget *widget;
gpointer data;
{
  DisplayReportData *drd = (DisplayReportData *)data;
  gtk_widget_destroy ( drd->filesel );
  drd->filesel = NULL;
}

static void display_results_save_ok_callback ( widget, data )
GtkWidget *widget;
gpointer data;
{
  DisplayReportData *drd = (DisplayReportData *)data;
  char *file;
  int fd;
  char *msg;

  file = gtk_file_selection_get_filename (
    GTK_FILE_SELECTION ( drd->filesel ) );
  fd = open ( file, O_WRONLY | O_CREAT, 0644 );
  if ( fd <= 0 ) {
    msg = (char *) malloc ( strlen ( file ) + 100 );
    sprintf ( msg, "Error %d writing to file\n%s",
      errno, file );
    create_confirm_window ( CONFIRM_ERROR,
      "GTimer: Error", msg, "Ok", NULL, NULL, NULL, NULL );
    free ( msg );
  } else {
    write ( fd, drd->text, strlen ( drd->text ) );
    close ( fd );
    gtk_widget_destroy ( drd->filesel );
    drd->filesel = NULL;
  }
}

static void display_results_save_callback ( widget, data )
GtkWidget *widget;
gpointer data;
{
  DisplayReportData *drd = (DisplayReportData *)data;

  if ( drd->filesel )
    return;

  drd->filesel = gtk_file_selection_new ( "TaskTime: Save Report" );

  gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION(drd->filesel)->ok_button),
    "clicked", GTK_SIGNAL_FUNC (display_results_save_ok_callback), drd);
  gtk_signal_connect (
    GTK_OBJECT (GTK_FILE_SELECTION(drd->filesel)->cancel_button),
    "clicked", GTK_SIGNAL_FUNC (display_results_save_cancel_callback), drd);

  gtk_widget_show ( drd->filesel );
}




static void display_results ( text )
char *text;
{
  DisplayReportData *drd;
  GtkWidget *window, *hbox, *button, *textarea, *vscrollbar;
  GtkStyle *style;
  /*GtkTooltips *tooltips;*/

  drd = (DisplayReportData *) malloc ( sizeof ( DisplayReportData ) );
  memset ( drd, '\0', sizeof ( DisplayReportData ) );
  drd->text = text;

  drd->window = window = gtk_dialog_new ();
  gtk_window_set_title (GTK_WINDOW (window), "GTimer: Report" );
  gtk_window_set_wmclass ( GTK_WINDOW ( drd->window ), "GTimer", "gtimer" );

  hbox = gtk_hbox_new ( FALSE, 2 );
  gtk_box_pack_start ( GTK_BOX ( GTK_DIALOG (window)->vbox ),
    hbox, TRUE, TRUE, 2 );

  style = gtk_style_new ();
  gdk_font_unref ( style->font );
  style->font = gdk_font_load (
    "-adobe-courier-medium-r-*-*-12-*-*-*-*-*-*-*" );
  if ( style->font == NULL )
    style->font = gdk_font_load ( "fixed" );
  gtk_widget_push_style ( style );

  textarea = gtk_text_new ( NULL, NULL );
  gtk_widget_set_usize ( textarea, 300, 200 );
  gtk_box_pack_start ( GTK_BOX ( hbox ),
    textarea, TRUE, TRUE, 2 );
  gtk_widget_show ( textarea );
  gtk_widget_realize ( textarea );
  gtk_text_freeze ( GTK_TEXT ( textarea ) );
  gtk_text_set_point ( GTK_TEXT ( textarea ), 0 );
  gtk_text_insert ( GTK_TEXT ( textarea ), NULL, NULL, NULL,
    text, strlen ( text ) );
  gtk_text_thaw ( GTK_TEXT ( textarea ) );

  gtk_widget_pop_style ();

  vscrollbar = gtk_vscrollbar_new ( GTK_TEXT ( textarea )->vadj );
  gtk_box_pack_start ( GTK_BOX ( hbox ),
    vscrollbar, FALSE, FALSE, 2 );
  gtk_widget_show ( vscrollbar );

  gtk_widget_show ( hbox );

  button = gtk_button_new_with_label ("Save");
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (window)->action_area),
    button, TRUE, FALSE, 10);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
    GTK_SIGNAL_FUNC (display_results_save_callback), drd);
  gtk_widget_show (button);

  button = gtk_button_new_with_label ("Ok");
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (window)->action_area),
    button, TRUE, FALSE, 10);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
    GTK_SIGNAL_FUNC (display_results_ok_callback), drd);
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_widget_grab_default (button);
  gtk_widget_show (button);

  gtk_widget_show ( window );
}



static time_t summarize_day ( fp, seltasks, num_seltasks, then, type,
  include_hours, include_annotations, is_last )
FILE *fp;
ReportTaskData **seltasks;
int num_seltasks;
time_t then;
report_type type;
int include_hours;
int include_annotations;
int is_last;
{
  struct tm *tm;
  int found = 0;
  int loop, loop2;
  int mon, mday, year, wday;
  TaskTimeEntry *entry;
  TaskAnnotation **anns;
  int num_anns = 0;
  int h, m, s;
  static char *weekdays[] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri",
    "Sat" };
  time_t ret = 0;
  char indentation[30];

  tm = localtime ( &then );
  mon = tm->tm_mon + 1;
  mday = tm->tm_mday;
  year = tm->tm_year + 1900;
  wday = tm->tm_wday;

  if ( include_hours )
    strcpy ( indentation, "            " );
  else
    strcpy ( indentation, "  " );

  for ( loop = 0; loop < num_seltasks; loop++ ) {
    if ( seltasks[loop]->week_start[0] == '\0' || tm->tm_wday == 0 )
      sprintf ( seltasks[loop]->week_start, "%02d/%02d/%d %s", mon, mday,
        year, weekdays[wday] );
    if ( include_hours )
      entry = taskGetTimeEntry ( seltasks[loop]->taskdata->task,
        year, mon, mday );
    else
      entry = NULL;
    if ( include_annotations ) {
      anns = TaskGetAnnotationEntries ( seltasks[loop]->taskdata->task,
        year, mon, mday, midnight_offset, &num_anns );
      for ( loop2 = 0; loop2 < num_anns; loop2++ ) {
        if ( type == REPORT_TYPE_DAILY )
          concat_annotation ( &seltasks[loop]->todays_annotations,
            anns[loop2], 0, 0, indentation );
        else if ( type == REPORT_TYPE_WEEKLY )
          concat_annotation ( &seltasks[loop]->weekly_annotations,
            anns[loop2], 0, 0, indentation );
        else if ( type == REPORT_TYPE_MONTHLY )
          concat_annotation ( &seltasks[loop]->monthly_annotations,
            anns[loop2], 0, 0, indentation );
        else if ( type == REPORT_TYPE_YEARLY )
          concat_annotation ( &seltasks[loop]->yearly_annotations,
            anns[loop2], 0, 0, indentation );
        else if ( type == REPORT_TYPE_TOTAL )
          concat_annotation ( &seltasks[loop]->total_annotations,
            anns[loop2], 0, 0, indentation );
      }
      if ( anns )
        free ( anns );
    }
    if ( ( entry && entry->seconds ) || seltasks[loop]->todays_annotations ) {
      if ( type == REPORT_TYPE_DAILY ) {
        if ( ! found ) {
          fprintf ( fp, "\n%02d/%02d/%d %s\n", mon, mday, year,
            weekdays[wday] );
          fprintf ( fp, "-------------------------------------\n" );
          found = 1;
        }
        if ( entry && entry->seconds ) {
          h = entry->seconds / 3600;
          m = ( entry->seconds - h * 3600 ) / 60;
          s = entry->seconds % 60;
        } else {
          h = m = s = 0;
        }
        if ( include_hours )
          fprintf ( fp, "%3d:%02d:%02d - %s\n", h, m, s,
            seltasks[loop]->taskdata->task->name );
        else
          fprintf ( fp, "%s\n", seltasks[loop]->taskdata->task->name );
      }
      if ( entry && entry->seconds ) {
        ret += entry->seconds;
        /* update totals */
        seltasks[loop]->weekly_total += entry->seconds;
        seltasks[loop]->monthly_total += entry->seconds;
        seltasks[loop]->yearly_total += entry->seconds;
        seltasks[loop]->total += entry->seconds;
      }
    }
    if ( seltasks[loop]->todays_annotations ) {
      fprintf ( fp, "%s", seltasks[loop]->todays_annotations );
      free ( seltasks[loop]->todays_annotations );
      seltasks[loop]->todays_annotations = NULL;
    }
  }

  if ( type == REPORT_TYPE_DAILY && ret ) {
    h = ret / 3600;
    m = ( ret - h * 3600 ) / 60;
    s = ret % 60;
    fprintf ( fp, "------\n" );
    fprintf ( fp, "%3d:%02d:%02d - %s\n", h, m, s, "Total" );
  }

  if ( ( type == REPORT_TYPE_WEEKLY ) && 
    ( wday == 6 || is_last ) ) {
    found = 0;
    for ( loop = 0; loop < num_seltasks; loop++ ) {
      if ( seltasks[loop]->weekly_total ||
        seltasks[loop]->weekly_annotations ) {
        if ( ! found ) {
          fprintf ( fp, "\nWeek %s to %02d/%02d/%d %s\n",
            seltasks[loop]->week_start, mon, mday, year,
            weekdays[wday] );
          fprintf ( fp, "-------------------------------------\n" );
          found = 1;
        }
        if ( include_hours ) {
          h = seltasks[loop]->weekly_total / 3600;
          m = ( seltasks[loop]->weekly_total - h * 3600 ) / 60;
          s = seltasks[loop]->weekly_total % 60;
          fprintf ( fp, "%3d:%02d:%02d - %s\n", h, m, s,
            seltasks[loop]->taskdata->task->name );
          seltasks[loop]->weekly_total = 0;
        } else
          fprintf ( fp, "%s\n", seltasks[loop]->taskdata->task->name );
        if ( seltasks[loop]->weekly_annotations ) {
          fprintf ( fp, "%s", seltasks[loop]->weekly_annotations );
          free ( seltasks[loop]->weekly_annotations );
          seltasks[loop]->weekly_annotations = NULL;
        }
      }
    }
  }

  if ( ( type == REPORT_TYPE_MONTHLY ) &&
    ( is_last ||
    ( ( ( tm->tm_year % 4 == 0 ) &&
      ( lmonth_days[tm->tm_mon] == tm->tm_mday ) ) ||
    ( ( tm->tm_year % 4 != 0 ) &&
      ( month_days[tm->tm_mon] == tm->tm_mday ) ) ) ) ) {
    found = 0;
    for ( loop = 0; loop < num_seltasks; loop++ ) {
      if ( seltasks[loop]->monthly_total ||
        seltasks[loop]->monthly_annotations ) {
        if ( ! found ) {
          fprintf ( fp, "\nMonth %02d/%d\n", mon, year );
          fprintf ( fp, "-------------------------------------\n" );
          found = 1;
        }
        if ( include_hours ) {
          h = seltasks[loop]->monthly_total / 3600;
          m = ( seltasks[loop]->monthly_total - h * 3600 ) / 60;
          s = seltasks[loop]->monthly_total % 60;
          fprintf ( fp, "%3d:%02d:%02d - %s\n", h, m, s,
            seltasks[loop]->taskdata->task->name );
          seltasks[loop]->monthly_total = 0;
        }
        else
          fprintf ( fp, "%s\n", seltasks[loop]->taskdata->task->name );
        if ( seltasks[loop]->monthly_annotations ) {
          fprintf ( fp, "%s", seltasks[loop]->monthly_annotations );
          free ( seltasks[loop]->monthly_annotations );
          seltasks[loop]->monthly_annotations = NULL;
        }
      }
    }
  }

  if ( ( type == REPORT_TYPE_YEARLY ) &&
    ( is_last || ( tm->tm_mon == 12 && tm->tm_mday == 31 ) ) ) {
    found = 0;
    for ( loop = 0; loop < num_seltasks; loop++ ) {
      if ( seltasks[loop]->yearly_total ||
        seltasks[loop]->yearly_annotations ) {
        if ( ! found ) {
          fprintf ( fp, "\nYear %d\n", year );
          fprintf ( fp, "-------------------------------------\n" );
          found = 1;
        }
        if ( include_hours ) {
          h = seltasks[loop]->yearly_total / 3600;
          m = ( seltasks[loop]->yearly_total - h * 3600 ) / 60;
          s = seltasks[loop]->yearly_total % 60;
          fprintf ( fp, "%3d:%02d:%02d - %s\n", h, m, s,
            seltasks[loop]->taskdata->task->name );
          seltasks[loop]->yearly_total = 0;
        }
        else
          fprintf ( fp, "%s\n", seltasks[loop]->taskdata->task->name );
        if ( seltasks[loop]->yearly_annotations ) {
          fprintf ( fp, "%s", seltasks[loop]->yearly_annotations );
          free ( seltasks[loop]->yearly_annotations );
          seltasks[loop]->yearly_annotations = NULL;
        }
      }
    }
  }

  if ( ( type == REPORT_TYPE_TOTAL ) && is_last ) {
    found = 0;
    for ( loop = 0; loop < num_seltasks; loop++ ) {
      if ( seltasks[loop]->yearly_total ) {
        if ( ! found ) {
          fprintf ( fp, "\nTotals\n" );
          fprintf ( fp, "-------------------------------------\n" );
          found = 1;
        }
        if ( include_hours ) {
          h = seltasks[loop]->total / 3600;
          m = ( seltasks[loop]->total - h * 3600 ) / 60;
          s = seltasks[loop]->total % 60;
          fprintf ( fp, "%3d:%02d:%02d - %s\n", h, m, s,
            seltasks[loop]->taskdata->task->name );
          seltasks[loop]->total = 0;
        }
        else
          fprintf ( fp, "%s\n", seltasks[loop]->taskdata->task->name );
        if ( seltasks[loop]->total_annotations ) {
          fprintf ( fp, "%s", seltasks[loop]->total_annotations );
          free ( seltasks[loop]->total_annotations );
          seltasks[loop]->total_annotations = NULL;
        }
      }
    }
  }

  if ( is_last ) {
    for ( loop = 0; loop < num_seltasks; loop++ ) {
      if ( seltasks[loop]->todays_annotations )
        free ( seltasks[loop]->todays_annotations );
      if ( seltasks[loop]->weekly_annotations )
        free ( seltasks[loop]->weekly_annotations );
      if ( seltasks[loop]->monthly_annotations )
        free ( seltasks[loop]->monthly_annotations );
      if ( seltasks[loop]->yearly_annotations )
        free ( seltasks[loop]->yearly_annotations );
      if ( seltasks[loop]->total_annotations )
        free ( seltasks[loop]->total_annotations );
      seltasks[loop]->todays_annotations = NULL;
      seltasks[loop]->weekly_annotations = NULL;
      seltasks[loop]->monthly_annotations = NULL;
      seltasks[loop]->yearly_annotations = NULL;
      seltasks[loop]->total_annotations = NULL;
    }
  }

  return ( ret );
}




static void ok_callback ( widget, data )
GtkWidget *widget;
gpointer data;
{
  ReportData *rd = (ReportData *) data;
  GList *selected, *item;
  int loop;
  ReportTaskData **seltasks;
  int num_selected = 0;
  int range = REPORT_RANGE_TODAY;
  FILE *fp;
  char *tempfile;
  time_t now, time_start, time_end, time_loop;
  struct tm *tm;
  struct stat buf;
  char *text = NULL;
  int fd, ret;

  selected = GTK_LIST ( rd->task_list ) ->selection;

  if ( ! selected ) {
    create_confirm_window ( CONFIRM_ERROR,
      "Error",
      "You have not selected any tasks",
      "Ok", NULL,
      NULL, NULL,
      NULL );
    return;
  }

  /* include hours / annotations*/
  rd->include_hours =
    ( GTK_TOGGLE_BUTTON ( rd->include_hours_checkbox ) -> active );
  rd->include_annotations =
    ( GTK_TOGGLE_BUTTON ( rd->include_annotations_checkbox ) -> active );

  if ( ! rd->include_hours && ! rd->include_annotations ) {
    create_confirm_window ( CONFIRM_ERROR,
      "Error",
      "You must select either \"include hours\"\nor \"include annotations\".",
      "Ok", NULL,
      NULL, NULL,
      NULL );
    return;
  }

  /* which tasks were selected... */
  seltasks = (ReportTaskData **) malloc
    ( sizeof ( ReportTaskData * ) * rd->num_tasks );
  for ( loop = 0; loop < rd->num_tasks; loop++ )
    seltasks[loop] = NULL;
  for ( item = selected; item != NULL; item = item->next ) {
    for ( loop = 0; loop < rd->num_tasks; loop++ ) {
      if ( item->data == rd->list_items[loop] ) {
        seltasks[num_selected] = (ReportTaskData *) malloc
          ( sizeof ( ReportTaskData ) );
        memset ( seltasks[num_selected], '\0', sizeof ( ReportTaskData ) );
        seltasks[num_selected++]->taskdata = rd->tasks[loop];
      }
    }
  }

  /* which time range ? */
  for ( loop = 0; time_options[loop]; loop++ ) {
    if ( GTK_OPTION_MENU ( rd->time_menu ) -> menu_item ==
      rd->time_menu_items[loop] ) {
      range = loop;
    }
  }

  gtk_widget_destroy ( rd->window );

  /* Generate the report...
  ** Write report to a file, then read it in...
  */
  tempfile = (char *) malloc ( 100 );
  sprintf ( tempfile, "/tmp/ttrpt-%d", (int) random () % 1000 );
  fp = fopen ( tempfile, "w" );
  if ( ! fp ) {
    create_confirm_window ( CONFIRM_ERROR,
      "Error",
      "Error opening temporary file",
      "Ok", NULL,
      NULL, NULL,
      NULL );
    free ( tempfile );
    for ( loop = 0; loop < num_tasks; loop++ )
      free ( seltasks[loop] );
    free ( seltasks );
    free ( rd->tasks );
    free ( rd->list_items );
    free ( rd );
    return;
  }
  time ( &now );
  now -= midnight_offset;
  tm = localtime ( &now );
  switch ( range ) {
    case REPORT_RANGE_TODAY:
      time_start = time_end = now;
      break;
    case REPORT_RANGE_LAST_WEEK:
      time_start = now - ONE_DAY * tm->tm_wday - ONE_DAY * 7;
      time_end = now - ONE_DAY * tm->tm_wday - ONE_DAY;
      break;
    case REPORT_RANGE_THIS_MONTH:
      time_start = now - ( ONE_DAY * ( tm->tm_mday - 1 ) );
      time_end = now;
      break;
    case REPORT_RANGE_LAST_MONTH:
      if ( tm->tm_mon == 1 )
        time_start = now - ( ONE_DAY * ( tm->tm_mday - 1 ) ) - ONE_DAY * 31;
      else if ( tm->tm_year % 4 == 0 )
        time_start = now - ( ONE_DAY * ( tm->tm_mday - 1 ) ) -
          ONE_DAY * lmonth_days[tm->tm_mon-1];
      else
        time_start = now - ( ONE_DAY * ( tm->tm_mday - 1 ) ) -
          ONE_DAY * month_days[tm->tm_mon-1];
      time_end = now - ONE_DAY * tm->tm_mday;
      break;
    case REPORT_RANGE_THIS_YEAR:
      time_start = now - ONE_DAY * tm->tm_yday;
      time_end = now;
      break;
    default:
    case REPORT_RANGE_THIS_WEEK:
      time_start = now - ONE_DAY * tm->tm_wday;
      time_end = now;
      break;
  }

  for ( time_loop = time_start; time_loop <= time_end; time_loop += ONE_DAY ) {
    summarize_day ( fp, seltasks, num_selected, time_loop, rd->type,
      rd->include_hours, rd->include_annotations, ( time_loop == time_end ) );
  }

  fclose ( fp );
  if ( stat ( tempfile, &buf ) == 0 ) {
    text = (char *) malloc ( buf.st_size + 1 );
    fd = open ( tempfile, O_RDONLY );
    ret = read ( fd, text, buf.st_size );
    close ( fd );
    if ( ret >= 0 )
      text[ret] = '\0';
    else
      sprintf ( text, "Error reading %s\n", tempfile );
    display_results ( text );
  }

  /* Free resources */
  unlink ( tempfile );
  free ( tempfile );
  for ( loop = 0; loop < num_selected; loop++ )
    free ( seltasks[loop] );
  free ( seltasks );
  free ( rd->tasks );
  free ( rd->list_items );
  free ( rd );
}


static void cancel_callback ( widget, data )
GtkWidget *widget;
gpointer data;
{
  ReportData *rd = (ReportData *) data;
  gtk_widget_destroy ( rd->window );
  free ( rd->tasks );
  free ( rd->list_items );
  free ( rd );
}





static GtkWidget *create_time_menu ( rd )
ReportData *rd;
{
  GtkWidget *menu;
  int loop;

  menu = gtk_menu_new ();

  for ( loop = 0; time_options[loop]; loop++ ) {
    rd->time_menu_items[loop] =
      gtk_menu_item_new_with_label ( time_options[loop] );
    gtk_menu_append ( GTK_MENU ( menu ), rd->time_menu_items[loop] );
    gtk_widget_show ( rd->time_menu_items[loop] );
  }
  return ( menu );
}



/*
** Create the report setup window.
** It's an add if taskdata is NULL.
*/
GtkWidget *create_report_window ( type )
report_type type;
{
  GtkWidget *report_window;
  GtkWidget *hbox;
  /*GtkTooltips *tooltips;*/
  GtkWidget *label, *ok_button, *cancel_button, *time_menu, *scrolled;
  ReportData *rd;
  GList *items = NULL;
  int loop;

  rd = (ReportData *) malloc ( sizeof ( ReportData ) );
  memset ( rd, '\0', sizeof ( ReportData ) );
  rd->type = type;
  rd->window = report_window = gtk_dialog_new ();
  gtk_window_set_wmclass ( GTK_WINDOW ( rd->window ), "GTimer", "gtimer" );
  gtk_window_set_title (GTK_WINDOW (report_window), "GTimer: Report" );
  gtk_window_position ( GTK_WINDOW(report_window), GTK_WIN_POS_MOUSE );
  gtk_grab_add ( report_window );
  gtk_widget_realize ( report_window );

  hbox = gtk_hbox_new ( TRUE, 5 );
  gtk_box_pack_start ( GTK_BOX ( GTK_DIALOG (report_window)->vbox ),
    hbox, TRUE, TRUE, 2 );
  gtk_widget_show ( hbox );

  label = gtk_label_new ( "Time Range: " );
  gtk_box_pack_start ( GTK_BOX ( hbox ), label, FALSE, FALSE, 2 );
  gtk_widget_show ( label );

  rd->time_menu = gtk_option_menu_new ();
  time_menu = create_time_menu ( rd );
  gtk_option_menu_set_menu ( GTK_OPTION_MENU ( rd->time_menu ), time_menu );
  gtk_option_menu_set_history ( GTK_OPTION_MENU ( rd->time_menu ),
    REPORT_RANGE_THIS_MONTH );
  gtk_box_pack_start (GTK_BOX ( hbox ), rd->time_menu, FALSE, FALSE, 2);
  gtk_widget_show ( rd->time_menu );

  rd->include_hours_checkbox = gtk_check_button_new_with_label (
    "Include time worked" );
  gtk_box_pack_start ( GTK_BOX ( GTK_DIALOG (report_window)->vbox ),
    rd->include_hours_checkbox, FALSE, FALSE, 10 );
  gtk_toggle_button_set_state (
    GTK_TOGGLE_BUTTON ( rd->include_hours_checkbox ), 1 );
  gtk_widget_show ( rd->include_hours_checkbox );
  
  rd->include_annotations_checkbox = gtk_check_button_new_with_label (
    "Include annotations" );
  gtk_box_pack_start ( GTK_BOX ( GTK_DIALOG (report_window)->vbox ),
    rd->include_annotations_checkbox, FALSE, FALSE, 10 );
  gtk_widget_show ( rd->include_annotations_checkbox );
  

  label = gtk_label_new ( "Tasks to include:" );
  gtk_label_set_justify ( GTK_LABEL ( label ), GTK_JUSTIFY_LEFT );
  gtk_box_pack_start ( GTK_BOX ( GTK_DIALOG (report_window)->vbox ),
    label, FALSE, FALSE, 2 );
  gtk_widget_show ( label );

  /* list of tasks */
  scrolled = gtk_scrolled_window_new ( NULL, NULL );
  gtk_widget_set_usize ( scrolled, 300, 200 );
  gtk_scrolled_window_set_policy ( GTK_SCROLLED_WINDOW (scrolled),
    GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC );
  gtk_box_pack_start ( GTK_BOX ( GTK_DIALOG (report_window)->vbox ),
    scrolled, TRUE, TRUE, 2 );
  gtk_widget_show ( scrolled );

  rd->task_list = gtk_list_new ();
  rd->tasks = (TaskData **) malloc ( sizeof ( TaskData * ) * num_tasks );
  rd->list_items = (GtkWidget **) malloc ( sizeof ( GtkWidget * ) * num_tasks );
  rd->num_tasks = num_tasks;
  for ( loop = 0; loop < num_tasks; loop++ ) {
    rd->list_items[loop] =
      gtk_list_item_new_with_label ( tasks[loop]->task->name );
    items = g_list_append ( items, rd->list_items[loop] );
    rd->tasks[loop] = tasks[loop];
    gtk_widget_show ( rd->list_items[loop] );
  }
  gtk_container_add (GTK_CONTAINER (scrolled), rd->task_list );
  gtk_list_append_items ( GTK_LIST ( rd->task_list ), items );
  gtk_list_set_selection_mode ( GTK_LIST ( rd->task_list ),
    GTK_SELECTION_MULTIPLE );
  for ( loop = 0; loop < num_tasks; loop++ ) {
    gtk_list_select_item ( GTK_LIST ( rd->task_list ), loop );
  }
  gtk_widget_show ( rd->task_list );
  
  /* add command buttons */
  /*tooltips = gtk_tooltips_new ();*/

  gtk_box_set_homogeneous ( GTK_BOX ( GTK_DIALOG (report_window)->action_area ),
    1 );
  ok_button = gtk_button_new_with_label ("Ok");
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (report_window)->action_area),
    ok_button, TRUE, FALSE, 10);
  gtk_signal_connect (GTK_OBJECT (ok_button), "clicked",
    GTK_SIGNAL_FUNC (ok_callback), rd);
  GTK_WIDGET_SET_FLAGS (ok_button, GTK_CAN_DEFAULT);
  gtk_widget_grab_default (ok_button);
  gtk_widget_show (ok_button);
  /*gtk_tooltips_set_tips (tooltips, ok_button,
    "Generate a report" );*/

  cancel_button = gtk_button_new_with_label ("Cancel");
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (report_window)->action_area),
    cancel_button, TRUE, FALSE, 10);
  gtk_signal_connect (GTK_OBJECT (cancel_button), "clicked",
    GTK_SIGNAL_FUNC (cancel_callback), rd);
  gtk_widget_show (cancel_button);
  /*gtk_tooltips_set_tips (tooltips, cancel_button,
    "Cancel without generating a report" );*/

  gtk_widget_show (report_window);

  return ( report_window );
}


